@@ -426,37 +426,28 @@ namespace transport
 				auto address = peer.router->GetNTCPAddress (!context.SupportsV6 ());
 				if (address && m_NTCPServer)
 				{
-#if BOOST_VERSION >= 104900
-					if (!address->host.is_unspecified ()) // we have address now
-#else
-					boost::system::error_code ecode;
-				  address->host.to_string (ecode);
-					if (!ecode)
-#endif
+					if (!peer.router->UsesIntroducer () && !peer.router->IsUnreachable ())
 					{
-						if (!peer.router->UsesIntroducer () && !peer.router->IsUnreachable ())
+						if(!m_NTCPServer->ShouldLimit())
 						{
-							if(!m_NTCPServer->ShouldLimit())
+							auto s = std::make_shared<NTCPSession> (*m_NTCPServer, peer.router);
+							if(m_NTCPServer->UsingProxy())
 							{
-								auto s = std::make_shared<NTCPSession> (*m_NTCPServer, peer.router);
-								if(m_NTCPServer->UsingProxy())
-								{
-									NTCPServer::RemoteAddressType remote = NTCPServer::eIP4Address;
-									std::string addr = address->host.to_string();
+								NTCPServer::RemoteAddressType remote = NTCPServer::eIP4Address;
+								std::string addr = address->host.to_string();
 
-									if(address->host.is_v6())
-										remote = NTCPServer::eIP6Address;
+								if(address->host.is_v6())
+									remote = NTCPServer::eIP6Address;
 
-									m_NTCPServer->ConnectWithProxy(addr, address->port, remote, s);
-								}
-								else
-									m_NTCPServer->Connect (address->host, address->port, s);
-								return true;
+								m_NTCPServer->ConnectWithProxy(addr, address->port, remote, s);
 							}
 							else
-							{
-								LogPrint(eLogWarning, "Transports: NTCP Limit hit falling back to SSU");
-							}
+								m_NTCPServer->Connect (address->host, address->port, s);
+							return true;
+						}
+						else
+						{
+							LogPrint(eLogWarning, "Transports: NTCP Limit hit falling back to SSU");
 						}
 					}
 				}
@@ -469,17 +460,8 @@ namespace transport
 				if (m_SSUServer && peer.router->IsSSU (!context.SupportsV6 ()))
 				{
 					auto address = peer.router->GetSSUAddress (!context.SupportsV6 ());
-#if BOOST_VERSION >= 104900
-					if (!address->host.is_unspecified ()) // we have address now
-#else
-					boost::system::error_code ecode;
-					address->host.to_string (ecode);
-					if (!ecode)
-#endif
-					{
-						m_SSUServer->CreateSession (peer.router, address->host, address->port);
-						return true;
-					}
+					m_SSUServer->CreateSession (peer.router, address->host, address->port);
+					return true;
 				}
 			}
 			LogPrint (eLogInfo, "Transports: No NTCP or SSU addresses available");
@@ -570,6 +552,23 @@ namespace transport
 						m_SSUServer->CreateSession (router);		// no peer test
 				}
 			}
+			if (i2p::context.SupportsV6 ())
+			{
+				// try to connect to few v6 addresses to get our address back
+				for (int i = 0; i < 3; i++)
+				{
+					auto router = i2p::data::netdb.GetRandomSSUV6Router ();
+					if (router)
+					{
+						auto addr = router->GetSSUV6Address ();
+						if (addr)
+							m_SSUServer->GetServiceV6 ().post ([this, router, addr]
+							{
+								m_SSUServer->CreateDirectSession (router, { addr->host, (uint16_t)addr->port }, false);				
+							});
+					}
+				}
+			}
 		}
 		else
 			LogPrint (eLogError, "Transports: Can't detect external IP. SSU is not available");
